from typing import List
from dataclasses import dataclass, field

from ....objects import NovaDataclass
from ....objects.nova_player import NovaBasicPlayer

@dataclass(repr=False)
class NovaServer(NovaDataclass):
    data:dict = field(repr=False)

    id:str = field(init=False)
    name:str = field(init=False)
    display_name:str = field(init=False)
    server_count:str = field(init=False)
    max_preview_players:int = field(init=False)
    available:bool = field(init=False)
    player_preview:List[NovaBasicPlayer] = field(init=False)
    
    def __post_init__(self):
        super().__post_init__()

        self.id = self.get("id")
        self.name = self.get("name")
        self.display_name = self.get("display_name")
        self.server_count = self.get("server_count")
        self.max_preview_players = self.get("max_preview_players")
        self.available = self.get("available")
        self.player_preview = [NovaBasicPlayer(player) for player in self.get("player_preview")]


from __future__ import annotations

import requests

from .errors import *
from .. import config

from . import NovaAPI

class NovaCDN(NovaAPI):
    """Class used to interact with CDN server at ``https://novauniverse.net/cdn/``."""
    def __init__(self, endpoint: str = None, silence_endpoint_warning:bool=False):
        self.silence_endpoint_warning = silence_endpoint_warning

        super().__init__(endpoint)
        self.is_cdn_endpoint()

    def is_cdn_endpoint(self) -> bool:
        """Check if this endpoint link is from the CDN and warns you if it's not."""
        if not self.silence_endpoint_warning:
            if "novauniverse.net/api" in self.endpoint:
                self.logger.warn("You are using an API endpoint in the 'NovaCDN' class. Please only use api endpoints in the 'NovaAPI' class or else you'll run into many bugs.")
                return False
            else:
                return True
        return None

    def get(self) -> requests.Response:
        """Send a get request to that CDN endpoint."""

        if self.endpoint is None: raise NoEndpointPassed()
        if config.performance_mode is False: # Does online check if performance mode is not enabled.
            if self.is_online is False: raise FailedConnectivityCheck()

        self.logger.info(f"Sending cdn get request to '{self.endpoint}'...")
        response_file = self.__http_session.get(self.endpoint)

        if response_file.status_code == 200:
            self.logger.info(f"CDN get request of '{self.endpoint}' was successful!")
            return response_file
        else:
            # I think all the error messages in the CDN are spit out in plain text on HTML. I'm not sure though. (Create a github issue if there's an issue with this anywhere.)
            raise UnSuccessfulOperation(response_file.text)


from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Callable
from devgoldyutils import LoggerAdapter

from ..api import NovaAPI
from .. import nova_logger

class Event():
    def __init__(self, event_name:str) -> None:
        """A base class all NovaUniverse.py events inherited from."""
        self.name = event_name

        self.functions_to_trigger:list = []

        self.logger = LoggerAdapter(nova_logger, prefix=self.name)
    
    def add_function(self, func:Callable):
        """Used by ``EventClient`` to add functions to events."""
        self.functions_to_trigger.append(func)
        self.logger.debug(
            f"Added '{func.__name__}' function to '{self.name}' func trigger list."
        )

class EndpointEvent(ABC, Event):
    """Allows you to easily create events from NovaUniverse.py endpoints."""
    def __init__(self, event_name, endpoint:str|None):
        self.endpoint = endpoint
        """The event's endpoint. 🔗"""

        super().__init__(event_name)

    @abstractmethod
    def loop(self, data:dict|None) -> bool:
        """
        This method is called each ``EventClient`` 💖heartbeat if the event is in use.
        You will have to handle this to the accordance of your event.

        -------------------

        Returning ``True`` indicates to the ``EventClient`` that the data has changed and it can trigger the event.

        Returning ``False`` indicates to the ``EventClient`` that the data has not changed and it shouldn't trigger the event.
        """
        ...

    @abstractmethod
    def trigger_event(self) -> None:
        """This method is ran when ``NovaClient`` gets an indication that the data has changed from ``Event.loop``."""
        ...


# Import all events under this module.
# --------------------------------------------
from .client_ready import ClientReady
from .player_join import PlayerJoin
from .player_leave import PlayerLeave

class Events(Enum):
    """
    An enum class of all available events.

    ---------------

    ⭐ Example:
    -------------
    These can be used like so::

        client = EventClient()

        @client.on_event(Events.CLIENT_READY)
        def client_is_ready():
            print("Client is ready!")

        @client.on_event(Events.PLAYER_JOIN)
        def on_player_join(player:NovaOnlinePlayer):
            print(f"{player.username} joined {player.server_name}!")

        @client.on_event(Events.PLAYER_LEAVE)
        def on_player_join(player:NovaOnlinePlayer):
            print(f"{player.username} left {player.server_name}!")

        client.start()

    """

    CLIENT_READY = ClientReady
    """
    Triggers when NovaClient is ready.
    
    Used like so::

        @client.on_event(Events.CLIENT_READY)
        def client_is_ready():
            print("Client is ready!")

    """

    PLAYER_JOIN = PlayerJoin
    """
    Triggers each time a player joins any lobby on the Nova Universe network. Passes :py:meth:`~novauniverse.interfaces.stats.server.nova_online_player.NovaOnlinePlayer` object to function.
    
    Used like so::
    
        @client.on_event(Events.PLAYER_JOIN)
        def on_player_join(player: NovaOnlinePlayer):
            print(f"{player.username} joined {player.server_name}!")

    """

    PLAYER_LEAVE = PlayerLeave
    """
    Triggers each time a player leaves any lobby on the Nova Universe network. Passes :py:meth:`~novauniverse.interfaces.stats.server.nova_online_player.NovaOnlinePlayer` object to function.
    
    Used like so::
    
        @client.on_event(Events.PLAYER_LEAVE)
        def on_player_join(player: NovaOnlinePlayer):
            print(f"{player.username} left {player.server_name}!")
    
    """


import ast

from . import EndpointEvent
from ..interfaces.stats.server import NovaOnlinePlayer

from ..api.endpoints import Endpoints

from typing import Set

class PlayerLeave(EndpointEvent):
    """Triggers each time a player leaves any lobby on the Nova Universe network."""
    def __init__(self):
        super().__init__("player_leave", Endpoints.PLAYERS_ONLINE)

        self.__old_players:Set[str] = set()
        self.__who_left:NovaOnlinePlayer = None

    def loop(self, data:dict) -> bool:
        current_online_players = set([str(player_data) for player_data in data["players"]])

        for player in self.__old_players - current_online_players:
            self.__who_left = NovaOnlinePlayer(ast.literal_eval(player))
            self.__old_players = current_online_players
            return True
        
        self.__old_players = current_online_players
        return False

    def trigger_event(self):
        for function in self.functions_to_trigger:
            function(self.__who_left)


import ast

from . import EndpointEvent
from ..interfaces.stats.server import NovaOnlinePlayer

from ..api.endpoints import Endpoints

from typing import Set

class PlayerJoin(EndpointEvent):
    """Triggers each time a player joins any lobby on the Nova Universe network."""
    def __init__(self):
        super().__init__("player_join", Endpoints.PLAYERS_ONLINE)

        self.__old_players:Set[str] = set()
        self.__who_joined:NovaOnlinePlayer = None
        self.__first_run:bool = True

    def loop(self, data:dict) -> bool:
        current_online_players = set([str(player_data) for player_data in data["players"]])

        for player in current_online_players - self.__old_players:
            if not self.__first_run:
                self.__who_joined = NovaOnlinePlayer(ast.literal_eval(player))
                self.__old_players = current_online_players
                return True

        self.__first_run = False
        self.__old_players = current_online_players
        return False

    def trigger_event(self):
        for function in self.functions_to_trigger:
            function(self.__who_joined)


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

project = 'NovaUniverse.py'
copyright = '2023, Goldy'
author = 'Goldy'
version = "v2.0dev4"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.todo", "sphinx.ext.viewcode", "sphinx.ext.autodoc"]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_theme_options = {
    "dark_css_variables": {
        "color-brand-primary": "red",
        "color-brand-content": "#FFAEAE",
        "color-admonition-background": "#0d0404",
    },

    "light_css_variables": {
        "color-brand-primary": "red",
        "color-brand-content": "#CC3333",
        "color-admonition-background": "#ffaeae",
    }
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ['_static']
html_logo = "_static/logo.png"
html_favicon = "_static/logo.png"


from . import EndpointEvent

class ClientReady(EndpointEvent):
    """Triggers when NovaClient is ready."""
    def __init__(self):
        super().__init__("client_ready", endpoint=None)

        self.__ready = False

    def loop(self, data) -> bool:
        if self.__ready is False:
            return True

        return False
        
    def trigger_event(self):
        for function in self.functions_to_trigger:
            self.__ready = True
            function()


from __future__ import annotations

import requests
from devgoldyutils import LoggerAdapter

from .errors import *
from .endpoints import Endpoints

from .. import nova_logger
from ..info import PACKAGE_NAME_WITH_VER

from .. import config

cache_dict = {
    "http_session": None
}

class NovaAPI():
    """The main class that handles all requests to the web server at ``https://novauniverse.net/api/``. """
    def __init__(self, endpoint:str=None):
        self.endpoint:str = endpoint
        
        self.__http_session = cache_dict.get("http_session", None)

        if self.__http_session is None:
            self.__http_session = requests.Session()
            self.__http_session.headers["User-Agent"] = PACKAGE_NAME_WITH_VER
            
            cache_dict["http_session"] = self.__http_session

        self.logger = LoggerAdapter(nova_logger, prefix="NovaAPI")

    @property
    def is_online(self) -> bool:
        try:
            self.logger.debug("Checking if Nova Universe API is online...")

            response = self.__http_session.get(Endpoints.connectivity_check)
            success = response.json()["success"]
        except (requests.exceptions.RequestException, KeyError):
            return False

        if success:
            self.logger.debug("Yes API is Online!")
            return True
        else:
            return False
        

    def get(self) -> dict|list:
        """Send a get request to that endpoint."""

        if self.endpoint is None: raise NoEndpointPassed()
        if config.performance_mode is False: # Does online check if performance mode is not enabled.
            if self.is_online is False: raise FailedConnectivityCheck()

        self.logger.info(f"Sending get request to '{self.endpoint}'...")
        response_json = self.__http_session.get(self.endpoint).json()
        self.logger.debug(f"Data from request --> {response_json}")

        if isinstance(response_json, list):
            # If it's a list just return it. (The consistency in this api, smh lol)
            return response_json

        response_json:dict
        if response_json.get("success", True):
            self.logger.info(f"Get request of '{self.endpoint}' was successful!")

            # Log if response was cached.
            if response_json.get("cached"): self.logger.warning(f"This response from '{self.endpoint}' was indicated cached by the API!")

            return response_json
        else:
            # I'm getting message and error here because some endpoints don't error with a message key. (again the consistency in this api 🤬)
            raise UnSuccessfulOperation(response_json.get("message", response_json.get("error")))

from .cdn import NovaCDN


import requests
from ... import Server

def test_server_stats_data():
    data_from_wrapper = Server().get_stats().data
    data_from_raw_request = requests.get(url="https://api.novauniverse.net/v1/novauniverse_mc/stats/extended").json()

    data_from_wrapper["cached"] = None
    data_from_raw_request["cached"] = None

    data_from_wrapper["system"]["localtime"] = None
    data_from_raw_request["system"]["localtime"] = None

    assert data_from_raw_request == data_from_wrapper

def test_players_data():
    data_from_wrapper = Server().get_stats().players_data
    data_from_raw_request = requests.get(url="https://api.novauniverse.net/v1/novauniverse_mc/players/online").json()

    data_from_wrapper["cached"] = None
    data_from_raw_request["cached"] = None

    assert data_from_raw_request == data_from_wrapper


import requests
from ... import Discord

def test_discord_stats():
    data_from_wrapper = Discord().get_stats().data
    data_from_raw_request = requests.get(url="https://api.novauniverse.net/v1/discord/").json()

    assert data_from_raw_request == data_from_wrapper


from dataclasses import dataclass, field
from datetime import datetime

from typing import List

from .. import NovaDataclass
from .tournament_team import TournamentTeam
from .tournament_player import TournamentPlayer

@dataclass(repr=False)
class NovaBasicTournament(NovaDataclass):
    """Represents a base dataclass for every tournament on nova universe."""
    data:dict = field(repr=False)

    id:int = field(init=False)
    """Returns ID of tournament."""

    date:datetime = field(init=False)
    """Returns date the tournament took place."""

    display_name:str = field(init=False)
    """Returns the display name of the tournament."""

    winner_team_id:int = field(init=False)
    """Returns winner team id."""

    players:List[TournamentPlayer] = field(init=False)
    """Returns all the players in that tournament."""
    
    teams:List[TournamentTeam] = field(init=False)
    """Returns all the teams in that tournament."""

    # Added properties.
    # -------------------
    winner_team:TournamentTeam = field(init=False)
    """Returns the team that won."""

    def __post_init__(self):
        super().__post_init__()

        self.id = self.get("id")
        self.date = datetime.strptime(self.get("date"), "%Y-%m-%d")
        self.display_name = self.get("display_name")
        self.winner_team_id = self.get("winner_team_id")
        self.players = [TournamentPlayer(player_data) for player_data in self.get("players")]
        self.teams = [TournamentTeam(team_data, self.players) for team_data in self.get("teams")]
        self.winner_team = [team for team in self.teams if team.team_number == self.winner_team_id][0]


BASE_DOMAIN = "novauniverse.net"
BASE_API_URL = "https://api." + BASE_DOMAIN

# CDN
# --------
CDN_BASE_URL = "https://" + BASE_DOMAIN + "/cdn"

# APIS
# --------
LEGACY_API_URL = "https://" + BASE_DOMAIN + "/api"

V1_API_URL = BASE_API_URL + "/v1"


class Endpoints:
    """Class containing all Nova Universe API endpoints."""

    # Connectivity Check
    # --------------------
    connectivity_check = (
        V1_API_URL + "/connectivity_check"
    )


    # News
    # -------------
    # TODO: This endpoint is from the legacy api so it may be removed soon.
    NEWS_ALL = (
        LEGACY_API_URL + "/news/all"
    )
        
    NEWS_LATEST = (
        LEGACY_API_URL + "/news/latest"
    )


    # Stats
    # ------------
    STATS_EXTENDED = (
        V1_API_URL + "/novauniverse_mc/stats/extended"
    )

    STATS_DISCORD = (
        V1_API_URL + "/discord"
    )


    # Players
    # ---------
    PLAYERS_ONLINE = (
        V1_API_URL + "/novauniverse_mc/players/online"
    )


    # MCF
    # ------
    MCF_RESULT = (
        V1_API_URL + "/tournaments/mcf/result"
    )


    # Nova Games
    # -----------
    NOVA_GAMES_RESULT = (
        V1_API_URL + "/tournaments/nova_games/result"
    )
    
    # TODO: This only exists in the legacy api so it might be removed soon.
    NOVA_GAMES_LIVE_STATS = (
        LEGACY_API_URL + "/nova_games/live_stats"
    )
    
    # TODO: Another legacy api thing that might be removed in the future. 😥
    NOVA_GAMES_DYNAMIC_CONFIGURATION = (
        LEGACY_API_URL + "/nova_games/dynamic_configuration"
    )

class CDNEndpoints:
    """Class containing all Nova Universe CDN endpoints."""

    NOVA_GAMES_ICONS_TEAM = (
        CDN_BASE_URL + "/novagames/icons/team"
    )



from dataclasses import dataclass, field

from ....objects import Timestamp, NovaDataclass

@dataclass
class System(NovaDataclass):
    data:dict = field(repr=False)

    localtime:Timestamp = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        self.localtime = Timestamp(self.get("localtime"))


from __future__ import annotations

from ... import BasicInterface, NovaAPI, Endpoints

from typing import List, Any

from .server_info import ServerInfo
from .nova_online_player import NovaOnlinePlayer

class Server(BasicInterface):
    """
    The interface for NovaAPI's ``/stats/extended`` endpoint.
    Allows you to get all 🌐server stats/info and see which players are 🟢online.
    """
    def __init__(self):
        super().__init__()
        
        self.stats_extended_api = NovaAPI(Endpoints.STATS_EXTENDED)
        self.players_online_api = NovaAPI(Endpoints.PLAYERS_ONLINE)

    def get_stats(self) -> ServerInfo:
        """Get's and returns all server stats."""
        return ServerInfo(self.stats_extended_api.get(), self.players_online_api.get())
    
    def get_online_players(self) -> List[NovaOnlinePlayer]:
        """Returns list of just the online players."""
        return [NovaOnlinePlayer(player_data) for player_data in self.players_online_api.get()["players"]]


from __future__ import annotations

from .. import SearchInterface, SearchBy, Search, NovaAPI, Endpoints
from .mcf_tournament import MCFTournament
from ... import objects
from ...objects.order_by import OrderByNotSupported

from typing import List, Dict

class MCF(SearchInterface):
    """
    The interface for NovaAPI's ``/mcf`` endpoint.
    Allows you to get latest, get all and search for 🔥mcf tournaments.
    """
    def __init__(self):
        super().__init__(
            supports = [SearchBy.ID, SearchBy.NAME],
            keys = {
                SearchBy.ID: "id",
                SearchBy.NAME: "display_name"
            }
            
        )

        self.mcf_result_api = NovaAPI(Endpoints.MCF_RESULT)

    def search(self, query: Search | int | str) -> MCFTournament | None:
        """Search for an mcf tournament!"""
        return super().search(query, self.get_all())

    def get_all(self) -> List[MCFTournament]:
        """Returns all 🔥mcf tournament results from api in a list. Returns empty list if none."""
        return [MCFTournament(tournament_data) for tournament_data in self.mcf_result_api.get()]

    def get_latest(self) -> MCFTournament | None:
        """Returns the latest 🔥mcf tournament result from api. Returns None if there are no tournaments."""
        data = self.mcf_result_api.get()
        data.reverse()

        return (lambda uwu: None if uwu is [] else MCFTournament(uwu[0]))(data)

    def get_top_players(self, order_by: objects.OrderBy = 0, max_players: int = 15) -> List[objects.TournamentPlayer]:
        """
        Returns the top ranked players based on your parameters.
        
        This method returns a list of ``novauniverse.objects.tournaments.tournament_player.TournamentPlayer`` but with all scores and kills from previous tournaments combined.
        """
        players:Dict[str, objects.TournamentPlayer] = {}

        if isinstance(order_by, objects.OrderBy): order_by = order_by.value

        if order_by in [0, 1]:
            for player in [player for mcf in self.get_all() for player in mcf.players]:
                if player.uuid in players:
                    players[player.uuid].score += player.score
                    players[player.uuid].kills += player.kills
                else:
                    player.team_number = None; player.uid = None # We null these properties as they are irrelevant and are different in every tournament.
                    players[player.uuid] = player

            key = (lambda: (lambda x: x[1].score) if order_by == 0 else (lambda x: x[1].kills))()

            return [item[1] for item in sorted(players.items(), key=key, reverse=True)[:max_players]]
        
        raise OrderByNotSupported(order_by)


from ....objects import NovaDataclass
from dataclasses import dataclass, field

from .global_ import Global
from .player_preview import PlayerPreview
from .nova_server import NovaServer
from .system import System
from .nova_online_player import NovaOnlinePlayer

from typing import List

@dataclass
class ServerInfo(NovaDataclass):
    data:dict = field(repr=False)
    players_data:dict = field(repr=False)

    global_:Global = field(init=False)
    """Returns class that contains global ``player_count`` and ``server_count``."""
    player_count:int = field(init=False)
    """Returns global player count of Nova Universe."""
    server_count:int = field(init=False)
    """Returns global server count of Nova Universe."""
    player_preview:PlayerPreview = field(init=False)
    """Returns player preview."""
    servers:List[NovaServer] = field(init=False)
    """Returns all servers on the Nova Universe network."""
    online_players:List[NovaOnlinePlayer] = field(init=False)
    """Returns all players that are online as NovaBasicPlayer object."""
    cached:bool = field(init=False)
    """Returns whether this data is cached or not."""
    system:System = field(init=False)
    """Returns the system class that lets you see get details like the localtime."""

    def __post_init__(self):
        super().__post_init__()

        self.global_ = Global(self.get("global"))
        self.player_count = self.global_.player_count
        self.server_count = self.global_.server_count
        self.player_preview = PlayerPreview(self.get("player_preview"))
        self.servers = [NovaServer(server_data) for server_data in self.get("servers")]
        self.online_players = [NovaOnlinePlayer(player_data) for player_data in self.get("players", data=self.players_data, default_value=[])]
        self.cached = self.get("cached")
        self.system = System(self.get("system"))


from __future__ import annotations

from dataclasses import dataclass, field

from ..nova_player import NovaBasicPlayer

@dataclass(repr=False)
class TournamentPlayer(NovaBasicPlayer):
    """A base tournament player class, used in interfaces like mcf and nova games because player data is very similar between them."""
    uid:int = field(init=False)
    """Returns uid of tournament player."""
    team_number:int = field(init=False)
    """Returns the team number of the player."""
    score:int = field(init=False)
    """Returns the score the player currently achieved."""
    kills:int = field(init=False)
    """Returns the current amount of kills the player achieved."""

    def __post_init__(self) -> None:
        super().__post_init__()

        self.uid = self.get("uid")
        self.team_number = self.get("team_number")
        self.score = self.get("score", default_value = 0)
        self.kills = self.get("kills", default_value = 0)


from dataclasses import dataclass, field

from .tournament_player import TournamentPlayer
from ..nova_dataclass import NovaDataclass

from typing import Tuple, List

@dataclass(repr=False)
class TournamentTeam(NovaDataclass):
    """Represents a team from a tournament."""
    data:dict = field(repr=False)
    players_data:List[TournamentPlayer] = field(repr=False)

    team_number:int = field(init=False)
    team_score:int = field(init=False)

    # Added fields.
    # ---------------
    players:Tuple[TournamentPlayer] = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        self.team_number = self.get("team_number")
        self.team_score = self.get("team_score")

        # Added attributes
        # ------------------
        self.players = tuple([player for player in self.players_data if player.team_number == self.team_number])


"""
🐲 Nova Universe - A modern & maintained wrapper for the Nova Universe API written in Python.

Copyright (C) 2023 - Dev Goldy
"""

# Logging stuff
# ---------------
from .info import LOGGER_NAME
from devgoldyutils.logging import add_custom_handler, log

nova_logger = add_custom_handler(log.getLogger(LOGGER_NAME), level=log.WARNING)
"""
The python ``logging.Logger()`` class for NovaUniverse.py.

---------------

⭐Example:
------------

You can disable and also adjust the level of logging from the api wrapper like this::

    import logging
    nova_logger.setLevel(logging.DEBUG)

If your using the ``novauniverse.EventClient`` class, a shortcut for enabling debugging would be this::

    EventClient(debug=True)

If you want to completely disable logging like example remove warnings, you can do that with::

    nova_logger.setLevel(logging.NOTSET)

"""

# Configuration
# ---------------
from .configuration import Config
config = Config()
"""🐉 NovaUniverse.py config."""

# Endpoint Interfaces.
# ------------------------
#from .interfaces.news import News
from .interfaces.stats.discord import Discord
from .interfaces.stats.server import Server, NovaOnlinePlayer
from .interfaces.mcf import MCF
#from .interfaces.nova_games import NovaGames


# Events
# ------------------------
from .events import Events
from .event_client import EventClient

from .utils.search import Search

# Backend objects
from .objects.nova_player import NovaBasicPlayer
from .objects.timestamp import Timestamp
from .objects.tournaments import NovaBasicTournament, TournamentPlayer, TournamentTeam
from .objects.order_by import OrderBy

# Backend utils
from .api import Endpoints, NovaAPI, NovaCDN


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Dict
from devgoldyutils import LoggerAdapter

from .. import nova_logger, errors
from ..utils.search import Search, SearchBy

class Interface():
    """The base interface where all NovaUniverse.py endpoint interfaces inherit from."""
    def __init__(self) -> None:
        ...

class BasicInterface(Interface):
    """A basic interface, nothing more... nothing less..."""
    # TODO: Change this docstring lmao

    def __init__(self) -> None:
        super().__init__()


# TODO: Finish this off..
class SearchInterface(ABC, Interface):
    """Adds searching to a basic NU.PY interface. Use this to add searching functionality to your interfaces."""
    def __init__(self, supports: List[SearchBy], keys: Dict[SearchBy, str]) -> None:
        self.__supports = supports
        self.__keys = keys

        self.logger = LoggerAdapter(nova_logger, prefix="SearchInterface")
        super().__init__()

    @abstractmethod
    def search(self, query: Search | int | str, objects: List[object] = None) -> object | None:
        """It is recommended to use ``novauniverse.utils.search.Search()`` as a query but strings and integers will also work and be handled respectively."""

        if isinstance(query, str) or isinstance(query, int):
            if query.isnumeric():
                query = Search(id = query)
            else:
                query = Search(name = query)

        if not query.search_by in self.__supports:
            query.not_supported(self)

        self.logger.debug(f"Searching in '{self.__class__.__name__}' by '{query.search_by.name}' for '{query.get_query()}'...")
        # TODO: Add fuzzy search in the future.
        for object in objects:
            if object.__dict__[self.__keys[query.search_by]] == query.get_query():
                self.logger.info(f"Found {query.get_query()} by '{query.search_by.name}'.")
                return object

        return None


# Root Imports
# --------------
from ..api import NovaAPI, Endpoints


VER = "2.0.2"

VERSION = f"{VER}" # Full version string.

PACKAGE_NAME= "NovaUniverse.py"
PACKAGE_NAME_WITH_VER = f"{PACKAGE_NAME} ({VERSION})"

LOGGER_NAME = f"\u001b[31m{PACKAGE_NAME}\u001b[0m"
"""The name of the logger. You can use this to get the logger with ``logging.getLogger()``."""

COPYRIGHT = "\nCopyright (c) 2023-present (Dev Goldy)"


config_dict = {
    "performance_mode" : False
}

class Config():
    """NovaUniverse.py config."""

    @property
    def performance_mode(self) -> bool:
        """
        When enabled some unnecessary features are disabled in favour for performance.

        WARNING: This will make the api wrapper less stable and friendly. So watch out it may bite 😳🐻.
        """
        return config_dict["performance_mode"]

    @performance_mode.setter
    def performance_mode(self, x:bool):
        config_dict["performance_mode"] = x


from ....objects.nova_dataclass import NovaDataclass

from dataclasses import dataclass, field

@dataclass(repr=False)
class MemberCount(NovaDataclass):
    data:dict = field(repr=False)

    total:int = field(init=False)
    bots:int = field(init=False)
    members:int = field(init=False)

    def __post_init__(self):
        self.total = self.get("total")
        self.bots = self.get("bots")

        self.members = ( self.total - self.bots )
        super().__post_init__()

@dataclass(repr=False)
class DiscordStats(NovaDataclass):
    data:dict = field(repr=False)

    member_count:MemberCount = field(init=False)
    
    def __post_init__(self):
        self.member_count = MemberCount(self.get("member_count"))
        super().__post_init__()


from __future__ import annotations
from io import StringIO
from dataclasses import dataclass
from devgoldyutils import DictDataclass

from prettyprinter import cpprint, install_extras
install_extras(include=["dataclasses"])

from .. import config, nova_logger

@dataclass
class NovaDataclass(DictDataclass):
    """The root NovaUniverse.py class that all dataclasses inherited from."""
    def __post_init__(self):
        self.logger = nova_logger
        super().__post_init__()

    # BETTER dataclass representation, also it's coloured. 🌈
    # ---------------------------------------------------------
    def __repr__(self) -> str:
        if config.performance_mode: # I've heard pretty-printer can be a bit slow, so to save performance I've disabled it in performance mode.
            return super().__repr__()
        
        text_stream = StringIO()
        cpprint(self, stream=text_stream, depth=3, max_seq_len=1)
        text_stream.seek(0)
        return text_stream.read()


from ..errors import NovaError

# BASE EXCEPTION
#------------------
class NovaAPIError(NovaError):
    """Error raised when a known error occurs on the API."""
    def __init__(self, message:str) -> None:
        super().__init__(message)


# Other exceptions.
#---------------------

class NoEndpointPassed(NovaAPIError):
    """Raised when no endpoint is passed to NovaAPI."""
    def __init__(self) -> None:
        super().__init__("No endpoint was passed to NovaAPI(), an endpoint must be passed to use it's methods.")

class FailedConnectivityCheck(NovaAPIError):
    """Raised when failed to connect to Nova Universe API."""
    def __init__(self) -> None:
        super().__init__("Failed to establish a connection with Nova Universe API. Failed connect check. The API is probably offline. Blame Anton.")

class UnSuccessfulOperation(NovaAPIError):
    """Raised when Nova Universe API returns unsuccessful operation."""
    def __init__(self, message_from_api) -> None:
        super().__init__(f"That operation was returned as unsuccessful by the Nova Universe API.\n Error --> {message_from_api}")


from enum import Enum
from ..errors import NovaError

class OrderBy(Enum):
    """Used in top player methods from tournament interfaces."""
    SCORE = 0
    KILLS = 1

class OrderByNotSupported(NovaError):
    def __init__(self, order_by:OrderBy) -> None:
        super().__init__(f"This method does not support ordering by '{order_by.name}'.")


from dataclasses import dataclass, field

from ....objects.nova_player import NovaBasicPlayer

@dataclass(repr=False)
class NovaOnlinePlayer(NovaBasicPlayer):
    """A online Nova Universe player. WOW their actually ONLINE! That's CRAZY!"""
    server_name:str = field(init=False)
    """Returns name of the server the player is in."""
    server_type_id:str = field(init=False)
    """Returns type id of the server the player is in."""
    server_type_name:str = field(init=False)
    """Returns type name of the server the player is in."""
    server_type_display_name:str = field(init=False)
    """Returns type display name of the server the player is in."""

    def __post_init__(self) -> None:
        super().__post_init__()

        self.server_name = self.get("server_name")
        self.server_type_id = self.get("server_type_id")
        self.server_type_name = self.get("server_type_name")
        self.server_type_display_name = self.get("server_type_display_name")


import logging as log
from devgoldyutils import DevGoldyUtilsError

from . import nova_logger

class NovaError(DevGoldyUtilsError):
    """The base class of all NovaUniverse.py errors."""
    def __init__(self, message: str, logger: log.Logger = None):
        """Raises a NovaUniverse.py exception."""
        if logger is None:
            logger = nova_logger

        super().__init__(message, logger)


from dataclasses import dataclass, field
from . import NovaDataclass

@dataclass(repr=False)
class NovaBasicPlayer(NovaDataclass):
    """A very basic nova universe player class."""
    data:dict = field(repr=False)

    uuid:str = field(init=False)
    """Returns UUID of novauniverse player."""
    username:str = field(init=False)
    """Returns username of novauniverse player."""
    name:str = field(init=False)
    """Aliases of ``username``."""

    def __post_init__(self):
        super().__post_init__()

        self.uuid = self.get("uuid")
        self.username = self.get("username")

        self.name = self.username


from datetime import datetime
from dataclasses import dataclass, field

from . import NovaDataclass

@dataclass(repr=False)
class Timestamp(NovaDataclass):
    """A NovaUniverse API timestamp object."""
    data:dict = field(repr=False)

    date:datetime = field(init=False)
    """The date and time at Zeeraa's house right now. Wait what, why do you want to stalk his date and time."""
    timezone:str = field(init=False)
    timezone_type:int = field(init=False)

    def __post_init__(self):
        """Returns timestamp as python datetime object and more."""
        super().__post_init__()

        self.date = datetime.strptime(self.get("date"), "%Y-%m-%d %H:%M:%S.%f")
        self.timezone = self.get("timezone")
        self.timezone_type = self.get("timezone_type")


from __future__ import annotations
from typing import overload, NoReturn
from enum import Enum

from ..errors import NovaError

# SearchBy Enum class
# ---------------------
class SearchBy(Enum):
    ID = 1
    NAME = 2

# Errors
# --------
class SearchGotNoArgs(NovaError):
    def __init__(self) -> None:
        super().__init__("'Search()' class must have either id or name passed. Like --> Search(name='UwU Dev Goldy')")

class SearchNotCompletelySupported(NovaError):
    def __init__(self, searched_by, interface:object) -> None:
        super().__init__(f"Searching by '{searched_by}' not supported by '{interface.__class__.__name__}' interface/endpoint.")

# Search class
# ---------------
class Search():
    """A util that allows you to search within an interface by id or name if supported."""
    @overload
    def __init__(self, id:str|int) -> None:
        """Search by id."""
        ...
    
    @overload
    def __init__(self, name:str) -> None:
        """Search by name."""
        ...

    def __init__(self, id:str|int=None, name:str=None) -> None:
        self.id = id
        self.name = name

        self.search_by:SearchBy|None = None

        if self.name is not None:
            self.search_by = SearchBy.NAME
        if self.id is not None:
            self.search_by = SearchBy.ID
        
        if self.search_by is None:
            raise SearchGotNoArgs()

    def get_query(self) -> str|int|None:
        """Returns the query."""
        if self.search_by is SearchBy.ID:
            return self.id
        if self.search_by is SearchBy.NAME:
            return self.name
        return None

    def not_supported(self, interface: object) -> NoReturn:
        """Raises error to warn user this interface does not support searching by id/name."""
        raise SearchNotCompletelySupported(self.search_by.name, interface)


from dataclasses import dataclass, field
from typing import List

from ....objects import NovaDataclass
from ....objects.nova_player import NovaBasicPlayer

@dataclass(repr=False)
class PlayerPreview(NovaDataclass):
    data:dict = field(repr=False)

    max_preview_items:int = field(init=False)
    content:List[NovaBasicPlayer] = field(init=False)
    additional:int = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        self.max_preview_items = self.get("max_preview_items")
        self.content = [NovaBasicPlayer(player) for player in self.get("content")]
        self.additional = self.get("additional")


from __future__ import annotations

from ... import BasicInterface, NovaAPI, Endpoints
from .discord_stats import DiscordStats

class Discord(BasicInterface):
    """
    The interface for NovaAPI's ``/stats/discord`` endpoint.
    Allows you to get 🎮discord stats.
    """
    def __init__(self):
        super().__init__()
        
        self.api = NovaAPI(Endpoints.STATS_DISCORD)

    def get_stats(self) -> DiscordStats:
        """Get's and returns all server stats."""
        return DiscordStats(self.api.get())


from dataclasses import dataclass, field

from ....objects import NovaDataclass

@dataclass(repr=False)
class Global(NovaDataclass):
    data:dict = field(repr=False)

    player_count:int = field(init=False)
    """Returns global player count of Nova Universe."""
    server_count:int = field(init=False)
    """Returns global server count of Nova Universe."""
    
    def __post_init__(self):
        super().__post_init__()

        self.player_count = self.get("player_count")
        self.server_count = self.get("server_count")


from .nova_dataclass import NovaDataclass

from .timestamp import Timestamp
from .nova_player import NovaBasicPlayer
from .order_by import OrderBy, OrderByNotSupported

from .tournaments import NovaBasicTournament
from .tournaments.tournament_team import TournamentTeam
from .tournaments.tournament_player import TournamentPlayer


import requests
from .. import MCF

from dataclasses import asdict
from typing import List, Dict

def test_mcf_get_all():
    data_from_wrapper = [mcf.data for mcf in MCF().get_all()]
    data_from_raw_request = requests.get(url="https://api.novauniverse.net/v1/tournaments/mcf/result").json()

    assert data_from_raw_request == data_from_wrapper

def test_mcf_get_latest():
    data_from_wrapper = MCF().get_latest().data
    data_from_raw_request = requests.get(url="https://api.novauniverse.net/v1/tournaments/mcf/result").json()[-1]

    assert data_from_raw_request == data_from_wrapper

def test_mcf_get_top_players():
    data_from_wrapper = [asdict(player) for player in MCF().get_top_players()]
    
    count = 0
    for player in data_from_wrapper:
        del player["data"]; del player["logger"]; del player["name"]
        data_from_wrapper[count] = player

        count += 1

    data_from_raw_request: List[dict] = requests.get(url="https://api.novauniverse.net/v1/tournaments/mcf/result").json()

    players: Dict[str, dict] = {}

    for player in [player for mcf in data_from_raw_request for player in mcf["players"]]:
        if player["uuid"] in players:
            players[player["uuid"]]["score"] = player["score"] + players[player["uuid"]]["score"]
            players[player["uuid"]]["kills"] = player["kills"] + players[player["uuid"]]["kills"]
        else:
            player["team_number"] = None; player["uid"] = None # We null these properties as they are irrelevant and are different in every tournament.
            players[player["uuid"]] = player

    data_from_raw_request = [item[1] for item in sorted(players.items(), key=lambda x: x[1]["score"], reverse=True)[:15]]

    assert data_from_raw_request == data_from_wrapper


from __future__ import annotations

import sys
import time
from typing import List, Type, NoReturn, Dict, Callable
from devgoldyutils import LoggerAdapter

import logging
from . import nova_logger

from .api import NovaAPI
from .events import Event, EndpointEvent, Events

class EventClient():
    """A basic NU.PY client used to listen into events on the network."""
    def __init__(self, debug=False) -> None:
        self.__endpoint_events: Dict[str, List[EndpointEvent]] = {
            None: []
        }

        if debug:
            nova_logger.setLevel(logging.DEBUG)

        self.logger = LoggerAdapter(nova_logger, prefix="EventClient")

        self.__stop = None

    # Event shit
    # ------------
    def add_event(self, event: Type[Event]) -> None:
        """Registers event on the client and starts running it."""
        active_event = event()

        self.logger.debug(
            f"Adding the event '{active_event.name}'..."
        )

        if isinstance(active_event, EndpointEvent):
            try:
                self.__endpoint_events[active_event.endpoint].append(active_event)
            except KeyError:
                self.logger.debug(
                    f"The event endpoint '{active_event.endpoint}' doesn't exist so I'm creating one in the client's events dict."
                )
                self.__endpoint_events[active_event.endpoint] = []
                self.__endpoint_events[active_event.endpoint].append(active_event)
            
            self.logger.debug(f"Added '{active_event.name}' to client event list.")

        # TODO: Don't forget to handle other event types here when they are added.

        return None

    def get_event_instance(self, event: Event) -> Event | None:
        """Gets the active instance of this event from the registered events pool in ``EventClient``."""
        for endpoint in self.__endpoint_events:

            for active_event in self.__endpoint_events[endpoint]:

                if isinstance(active_event, event):
                    return active_event
        
        return None

    # Decorator shit
    # ---------------
    def on_event(self, event: Events) -> None:
        """Decorator that allows you to run a function when an event occurs on NovaUniverse network."""
        # I'm using issubclass() here instead of isinstance() because these are just references to a class and not an actual instance of a class. 
        # Hence isinstance() will not work for my case.
        if issubclass(event.value, EndpointEvent):
            if not event.value in [event.__class__ for endpoint in self.__endpoint_events for event in self.__endpoint_events[endpoint]]:
                self.add_event(event.value)

        # Handle other type of events here...

        def inner(func):
            # Add function to event trigger list.
            self.get_event_instance(event.value).add_function(func)

        return inner

    # Client stuff UwU.
    # ------------------
    def start(self) -> NoReturn:
        """Runs client and starts listening to events."""
        self.logger.info("Starting event loop and listening to events..."); self.__stop = False
        
        # Deee Loop, more like DEEZ NUTS!
        try:

            while self.__stop is not True:
                self.logger.debug("❤")
                
                for endpoint in self.__endpoint_events:
                    data = None

                    if not endpoint is None:
                        data = NovaAPI(endpoint).get()

                    for event in self.__endpoint_events[endpoint]:
                        if event.loop(data) is True: event.trigger_event()

                self.logger.debug("\x1b[31;20m" + "❤" + "\x1b[0m" + "\n")

                time.sleep(3)

        except KeyboardInterrupt:
            self.logger.info("Event Loop Stopped!")

    def stop(self):
        """Stops the client from running. 😏If you can even get to this method."""
        self.__stop = True